"""Fitting the asymptotic relation to an SNR spectrum

This module fits the asymptotic relation to the p-modes in a frequency range
around nu_max, the central frequency of the seismic mode envelope,
in a solar-like oscillator. Only l=0 and l=2 are fit, l=1 modes are ignored.
"""

import numpy as np
import pbjam as pb
import pandas as pd
import scipy.stats as scist
from .priors import kde
from .plotting import plotting

def get_nmax(dnu, numax, eps):
    """Compute radial order at numax.

    Compute the radial order at numax, which in this implimentation of the
    asymptotic relation is not necessarily integer.

    Parameters
    ----------
    numax : float
        Frequency of maximum power of the p-mode envelope (muHz).
    dnu : float
        Large separation of l=0 modes (muHz).
    eps : float
        Epsilon phase term in asymptotic relation (muHz).

    Returns
    -------
        nmax : float
            non-integer radial order of maximum power of the p-mode envelope
    """

    return numax / dnu - eps


def get_enns(nmax, norders):
    """Compute radial order numbers.

    Get the enns that will be included in the asymptotic relation fit. These
    are all integer.

    Parameters
    ----------
    nmax : float
        Frequency of maximum power of the p-mode envelope
    norders : int
        Total number of radial orders to consider

    Returns
    -------
    enns : ndarray
            Numpy array of norders radial orders (integers) around numax (nmax).
    """

    below = np.floor(nmax - np.floor(norders/2)).astype(int)
    above = np.floor(nmax + np.ceil(norders/2)).astype(int)
    
    # Handling of single input (during fitting), or array input when evaluating
    # the fit
    if type(below) != np.ndarray:
        return np.arange(below, above)
    else:
        return np.concatenate([np.arange(x, y) for x, y in zip(below, above)]).reshape(-1, norders)


def asymptotic_relation(numax, dnu, eps, alpha, norders):
    """ Compute the l=0 mode frequencies from the asymptotic relation for
    p-modes

    Parameters
    ----------
    numax : float
        Frequency of maximum power of the p-mode envelope (muHz).
    dnu : float
        Large separation of l=0 modes (muHz).
    eps : float
        Epsilon phase term in asymptotic relation (unitless).
    alpha : float
        Curvature factor of l=0 ridge (second order term, unitless).
    norders : int
        Number of desired radial orders to calculate frequncies for, centered
        around numax.

    Returns
    -------
    nu0s : ndarray
        Array of l=0 mode frequencies from the asymptotic relation (muHz).

    """
    nmax = get_nmax(dnu, numax, eps)
    enns = get_enns(nmax, norders)
    return (enns.T + eps + alpha/2*(enns.T - nmax)**2) * dnu


def P_envelope(nu, hmax, numax, width):
    """ Power of the seismic p-mode envelope

    Computes the power at frequency nu in the p-mode envelope from a Gaussian
    distribution. Used for computing mode heights.

    Parameters
    ----------
    nu : float
        Frequency (in muHz).

    hmax : float
        Height of p-mode envelope (in SNR).

    numax : float
        Frequency of maximum power of the p-mode envelope (in muHz).

    width : float
        Width of the p-mode envelope (in muHz).

    Returns
    -------
    h : float
        Power at frequency nu (in SNR)
    """
    hmax = 10**hmax
    width = 10**width
    return hmax * np.exp(- 0.5 * (nu - numax)**2 / width**2)


def get_summary_stats(fit, model, pnames):
    """ Make dataframe with fit summary statistics

    Creates a dataframe that contains various quantities that summarize the
    fit. Note, these are predominantly derived from the marginalized posteriors.

    Parameters
    ----------
    fit : asy_peakbag.mcmc instance
        asy_peakbag.mcmc that was used to fit the spectrum, containing the
        log-likelihoods and MCMC chains.
    model : the asymp_spec_model.model instance that defines the model used to
        fit the spectrum.
    pnames : list
       List of names of each of the parameters in the fit.

    Returns
    -------
    summary : pandas.DataFrame
        Dataframe with the summary statistics.
    mle_model : ndarray
        Numpy array with the model spectrum corresponding to the maximum
        likelihood solution.
    """

    summary = pd.DataFrame()
    idx = np.argmax(fit.flatlnlike)
    mle = fit.flatchain[idx,:]
    means = np.mean(fit.flatchain, axis = 0)
    stds = np.std(fit.flatchain, axis = 0)
    skewness = scist.skew(fit.flatchain, axis = 0)
    pars_percs = np.percentile(fit.flatchain, [50-95.4499736104/2,
                                               50-68.2689492137/2,
                                               50,
                                               50+68.2689492137/2,
                                               50+95.4499736104/2], axis=0)
    mads =  scist.median_absolute_deviation(fit.flatchain, axis=0)

    smry_stats = ['mle', 'mean', 'std', 'skew', '2nd', '16th', '50th', '84th',
                  '97th', 'MAD']
    for i, par in enumerate(pnames):
        z = [mle[i], means[i], stds[i], skewness[i],  pars_percs[0, i],
             pars_percs[1, i], pars_percs[2, i], pars_percs[3, i],
             pars_percs[4, i], mads[i]]
        A = {key: z[i] for i, key in enumerate(smry_stats)}
        summary[par] = pd.Series(A)
    summary = summary.transpose()
    mle_model = model(mle)
    return summary, mle_model


class asymp_spec_model():
    """Class for spectrum model using asymptotic relation.

    Parameters
    ---------_
    f : float, ndarray
        Array of frequency bins of the spectrum (muHz). Truncated to the range
        around numax.
    norders : int
        Number of radial order to fit

    Attributes
    ----------
    f : float, ndarray
        Array of frequency bins of the spectrum (muHz). Truncated to the range
        around numax.

    norders : int
        Number of radial order to fit
    """

    def __init__(self, f, norders):
        self.f = f
        self.norders = norders


    def lor(self, freq, h, w):
        """ Lorentzian to describe a mode.

        Parameters
        ----------
        freq : float
            Frequency of lorentzian (muHz).
        h : float
            Height of the lorentizan (SNR).
        w : float
            Full width of the lorentzian (log10(muHz)).

        Returns
        -------
        mode : ndarray
            The SNR as a function frequency for a lorentzian.
        """

        w = 10**(w)
        return h / (1.0 + 4.0/w**2*(self.f - freq)**2)


    def pair(self, freq0, h, w, d02, hfac=0.7):
        """Define a pair as the sum of two Lorentzians.

        A pair is assumed to consist of an l=0 and an l=2 mode. The widths are
        assumed to be identical, and the height of the l=2 mode is scaled
        relative to that of the l=0 mode. The frequency of the l=2 mode is the
        l=0 frequency minus the small separation.

        Parameters
        ----------
        freq0 : float
            Frequency of the l=0 (muHz).
        h : float
            Height of the l=0 (SNR).
        w : float
            The mode width (identical for l=2 and l=0) (log10(muHz)).
        d02 : float
            The small separation (muHz).
        hfac : float, optional
            Ratio of the l=2 height to that of l=0 (unitless).

        Returns
        -------
        pair_model : array
            The SNR as a function of frequency of a mode pair.
        """

        pair_model = self.lor(freq0, h, w)
        pair_model += self.lor(freq0 - d02, h*hfac, w)
        return pair_model

    def model(self, dnu, numax, eps, d02, alpha, hmax, envwidth, modewidth,
              *args):
        """ Constructs a spectrum model from the asymptotic relation.

        The asymptotic relation for p-modes with angular degree, l=0, is
        defined as:

        nu_nl = (n + epsilon + alpha/2(n - nmax)**2) * log(dnu) ,

        where nmax = numax / dnu - epsilon.

        We separate the l=0 and l=2 modes by the small separation d02.

        Parameters
        ----------
        dnu : float
            Large separation log10(muHz)
        lognumax : float
            Frequency of maximum power of the p-mode envelope log10(muHz)

        eps : float
            Phase term of the asymptotic relation (unitless)
        alpha : float
            Curvature of the asymptotic relation log10(unitless)
        d02 : float
            Small separation log10(muHz)
        loghmax : float
            Gaussian height of p-mode envelope log10(SNR)
        logenvwidth : float
            Gaussian width of the p-mode envelope log10(muHz)
        logmodewidth : float
            Width of the modes (log10(muHz))
        *args : array-like
            List of additional parameters (Teff, bp_rp) that aren't actually
            used to construct the spectrum model, but just for evaluating the
            prior.

        Returns
        -------
        model : ndarray
            spectrum model around the p-mode envelope
        """

        f0s = asymptotic_relation(10**numax, 10**dnu, eps, 10**alpha, self.norders)
        Hs = P_envelope(f0s, hmax, 10**numax, envwidth)
        mod = np.ones(len(self.f))
        for n in range(len(f0s)):
            mod += self.pair(f0s[n], Hs[n], modewidth, 10**d02)
        return mod

    def __call__(self, p):
        """ Produce model of the asymptotic relation

        Parameters
        ----------
        p : list
            list of model parameters

        Returns
        -------
        model : array
            spectrum model around the p-mode envelope
        """

        return self.model(*p)


class asymptotic_fit(kde, plotting):
    """ Class for fitting a spectrum based on the asymptotic relation.

    Parameters
    ----------
    f : ndarray
        Numpy array of frequency bins of the spectrum (muHz).
    s : ndarray
        Numpy array of power in each frequency bin (SNR).
    start_samples: ndarray
        Samples representing the starting guess for the asymptotic
        peakbagging.
    teff : [real, real]
        Stellar effective temperature and uncertainty
    bp_rp : [real, real]
        The Gaia Gbp - Grp color value and uncertainty (probably ~< 0.01 dex)
    store_chains : bool, optional
        Flag for storing all the full set of samples from the MCMC run.
        Warning, if running multiple targets, make sure you have enough memory.
    nthreads : int, optional
        Number of multiprocessing threads to use to perform the fit. For long
        cadence data 1 is best, more will just add parallelization overhead.
        Untested on short cadence.
    norders : int, optional
        Number of radial orders to fit

    Attributes
    ----------
    f : ndarray
        Numpy array of frequency bins of the spectrum (muHz).
    s : ndarray
        Numpy array of power in each frequency bin (SNR).
    sel : ndarray, bool
        Numpy array of boolean values specifying the frequency range to be
        considered in the asymptotic relation fit.
    model : asy_peakbag.model.model instance
        Function for computing a spectrum model given a set of parameters.
    bounds : ndarray
        Numpy array of upper and lower boundaries for the asymptotic relation
        fit. These limits truncate the likelihood function.
    gaussian : ndarray
        Numpy array of tuples of mean and sigma for Gaussian
        priors on each of the fit parameters (To be removed when full
        KDE is implimented).
    """

    def __init__(self, starinst, kdeinst=None, norders=6, store_chains=False, 
                 nthreads=1, path=None):
        
        self.pg = starinst.pg
        self.f = starinst.f
        self.s = starinst.s
        self.store_chains = store_chains
        self.nthreads = nthreads
        self.norders = norders
        
        self.par_names = ['dnu', 'numax', 'eps', 'd02', 'alpha', 'env_height',
                          'env_width', 'mode_width', 'teff', 'bp_rp']

        if kdeinst:
            self.start_samples = kdeinst.samples
            self.prior = kdeinst.prior
        elif hasattr(starinst, 'kde'):
            self.start_samples = starinst.kde.samples
            self.prior = starinst.kde.prior
        else:
            raise ValueError("Asy_peakbag won't run without samples of the prior")


        means = np.median(self.start_samples, axis=0)
        start = [10**(means[0]), 10**(means[1]), means[2], 10**(means[3]), 
                 10**(means[4]), means[5], means[6], means[7], 10**(means[8]),
                 means[9]]
        self.start = start

        nmax = get_nmax(*start[:3])
        enns = get_enns(nmax, self.norders)

        lower_frequency = (min(enns) - 1.25 + start[2]) * start[0]
        upper_frequency = (max(enns) + 0.25 + start[2]) * start[0]

        self.sel = (lower_frequency < self.f) & (self.f < upper_frequency)

        self.model = asymp_spec_model(self.f[self.sel], self.norders)

        self.modeID = None
        self.summary = None
        self.samples = None
        self.lnlike_fin = None
        self.lnprior_fin = None
        self.mle_model = None
        self.acceptance = None
                
        starinst.asy_fit = self


    def get_modeIDs(self, fit, N):
        """ Set mode ID in a dataframe

        Evaluates the asymptotic relation for each walker position from the
        MCMC fit. The median values of the resulting set of frequencies are
        then returned in a pandas.DataFrame

        Parameters
        ----------
        fit : asy_peakbag.mcmc class instance
            mcmc class instances used in the fit
        N : int
            Number of radial orders to output. Note that doesn't have to be
            the same as that used int he fit itself.

        Returns
        -------
        modeID : pandas.DataFrame
            Dataframe of radial order, n (best guess), angular degree, l,
            frequency and frequency error.
        """

        flatchain = fit.flatchain

        nsamps = np.shape(flatchain)[0]

        nu0_samps, nu2_samps = np.empty((nsamps, N)), np.empty((nsamps, N))

        nu0_samps = asymptotic_relation(10**flatchain[:, 1], 10**flatchain[:, 0],
                                        flatchain[:, 2], 10**flatchain[:, 4], N)
        nu2_samps = nu0_samps - 10**flatchain[:, 3]

        nus_med = np.median(np.array([nu0_samps, nu2_samps]), axis=2)
        nus_mad = scist.median_absolute_deviation(np.array([nu0_samps, nu2_samps]), axis=2)

        ells = [0 if i % 2 else 2 for i in range(2*N)]

        nus_med_out = []
        nus_mad_out = []

        for i in range(N):
            nus_med_out += [nus_med[1, i], nus_med[0, i]]
            nus_mad_out += [nus_mad[1, i], nus_mad[0, i]]

        modeID = pd.DataFrame({'ell': ells, 'nu_med': nus_med_out, 
                               'nu_mad': nus_mad_out})
        return modeID



    def likelihood(self, p):
        """ Likelihood function for set of model parameters

        Evaluates the likelihood function for a set of
        model parameters.  This includes the constraint from
        the observed variables.

        Parameters
        ----------
        p : array
            Array of model parameters

        Returns
        -------
        like : float
            likelihood function at p

        Note:
        log_dnu, log_numax, eps, log_d02, log_alpha, \
            log_env_height, log_env_width, log_mode_width, \
            log_teff, bp_rp = p
        """
        
        # Constraint from input obs
        ld = 0.0
        ld += self.normal(p[-2], *self.log_obs['teff'])
        ld += self.normal(p[-1], *self.log_obs['bp_rp'])
        
        # Constraint from the periodogram
        mod = self.model(p)
        lnlike = -1.0 * np.sum(np.log(mod) + self.s[self.sel] / mod)
        return lnlike + ld

    def __call__(self, dnu=[1, -1], numax=[1, -1], teff=[1, -1], bp_rp=[1, -1]):
        """ Setup, run and parse the asymptotic relation fit using EMCEE.

        Parameters
        ----------
        dnu : [real, real]
            Large frequency spacing and uncertainty
        numax : [real, real]
            Frequency of maximum power and uncertainty
        teff : [real, real]
            Stellar effective temperature and uncertainty
        bp_rp : [real, real]
            The Gaia Gbp - Grp color value and uncertainty 
            (probably ~< 0.01 dex).

        Returns
        -------
        asy_result : Dict
            A dictionary of the modeID DataFrame and the summary DataFrame.
        """
        
        self.obs = {'dnu': dnu, 'numax': numax, 'teff': teff, 'bp_rp': bp_rp}
        self.obs_to_log(self.obs)

        
        self.fit = pb.mcmc(np.median(self.start_samples, axis=0), self.likelihood,
                           self.prior, nthreads=self.nthreads)

        self.fit(start_samples=self.start_samples)

        self.modeID = self.get_modeIDs(self.fit, self.norders)

        self.summary, self.mle_model = get_summary_stats(self.fit, self.model, 
                                                         self.par_names)

        #if self.store_chains:
        self.samples = self.fit.flatchain
        #    self.lnlike_fin = self.fit.flatlnlike
        #else:
        #    self.samples = self.fit.chain[:,-1,:]
        #    self.lnlike_fin = np.array([self.fit.likelihood(self.fit.chain[i,-1,:]) for i in range(self.fit.nwalkers)])
        #    self.lnprior_fin = np.array([self.fit.lp(self.fit.chain[i,-1,:]) for i in range(self.fit.nwalkers)])

        self.acceptance = self.fit.acceptance
        
        return {'modeID': self.modeID, 'summary': self.summary}


from . import PACKAGEDIR
import os
import numpy as np

def get_priorpath():
    return os.path.join(*[PACKAGEDIR, 'data', 'prior_data.csv'])


def get_percentiles(X, sigma = 2, **kwargs):
    percs = np.array([0.682689492137,
                      0.954499736104,
                      0.997300203937,
                      0.999936657516,
                      0.999999426697,
                      0.999999998027])*100/2    
    percs = np.append(0, percs)    
    percs = np.append(-percs[::-1][:-1],percs)
    percs += 50
    print(kwargs)
    return np.percentile(X, percs[6-sigma : 6+sigma+1], **kwargs)


import numpy as np
import pandas as pd
from .mcmc import mcmc
from . import PACKAGEDIR
import os
from .plotting import plotting
import warnings
import statsmodels.api as sm


def default_priorpath():
    return os.path.join(*[PACKAGEDIR, 'data', 'prior_data.csv'])


class kde(plotting):
    ''' A class to predict epsilon.

    TODO: See the docs for full information. (especially example_advanced)

    '''
    def __init__(self, starinst=None, nthreads=1, verbose=False, bw_fac=1):

        if starinst:
            self.prior_file = starinst.prior_file
            self.f = starinst.f
            self.s = starinst.s
            self.pg = starinst.pg
            starinst.kde = self
        else:
            self.prior_file = default_priorpath()

        self.verbose = verbose
        self.nthreads = nthreads
        self.obs = []
        self.samples = []
        self.par_names = []
        self.bw_fac = bw_fac

        self.read_prior_data()

    def read_prior_data(self):
        ''' Read in the prior data from self.data_file '''
        self.prior_data = pd.read_csv(self.prior_file)
        #self.prior_data = self.prior_data.dropna()

    def select_prior_data(self, numax=None, nsigma=1):
        ''' Selects only the useful prior data based on proximity to estimated
            numax.

        Inputs
        ------
        numax: length 2 list [numax, numax_err]
            The estimate of numax together with uncertainty in log space.
            If numax==None then no selection will be made - all data will
            be used (you probably don't want to do this).

        '''
        if not numax:
            return self.prior_data

        # If the number of targets in the range considered for the prior is 
        # less than 100, the range will be expanded until it ~100. This is 
        # to ensure that the KDE can be constructed. Note: does not ensure
        # that the KDE is finite at the location of your target
       
        idx = np.zeros(len(self.prior_data), dtype = bool)
        while len(self.prior_data[idx]) < 100:
            nsigma += 0.5
            idx = np.abs(self.prior_data.numax.values - numax[0]) < nsigma * numax[1]
            if nsigma > 100:
                break
            
        if len(self.prior_data[idx]) > 1000:
            warnings.warn('You have lots data points in your prior - estimating' +
                          ' the KDE band width will be slow!')

        self.prior_data = self.prior_data[idx]

    def make_kde(self):
        ''' Takes the prior data and constructs a KDE function

        TODO: add details on the band width determination - see example
        advanced for a base explaination.

        I have settled on using the values from a cross validated maximum
        likelihood estimate.

        '''
        self.par_names = ['dnu', 'numax', 'eps', 'd02', 'alpha', 'env_height',
                          'env_width', 'mode_width', 'teff', 'bp_rp']

        self.select_prior_data(self.log_obs['numax'])
        
        if self.verbose:
                print(f'Selected data set length {len(self.prior_data)}')
                
        if self.bw_fac != 1:    
            from statsmodels.nonparametric.bandwidths import select_bandwidth
            bw = select_bandwidth(self.prior_data[self.par_names].values, 
                                  bw = 'scott', 
                                  kernel=None) * self.bw_fac
        else:
            if self.verbose:
                print('Selecting sensible stars ... for kde')
                print(f'Full data set length {len(self.prior_data)}')
            bw = 'cv_ml'

        self.kde = sm.nonparametric.KDEMultivariate(
                            data=self.prior_data[self.par_names].values,
                            var_type='cccccccccc', bw=bw)

    def normal(self, y, mu, sigma):
        ''' Returns normal log likelihood

        Inputs
        ------
        y : real
            observed value
        mu : real
            distribution mean
        sigma : real
            distribution standard deviation

        Returns
        -------
        log likelihood : real
        '''
        if (sigma < 0):
            return 0.0
        return -0.5 * (y - mu)**2 / sigma**2

    def prior(self, p):
        ''' Calculates the log prior from the KDE for the parameters p and
        applies some boundaries.

        Inputs
        ------
        p : array
            Array of the parameters

        Returns
        -------
        like : real
            The log likelihood evaluated at p.

        Note: p = ['dnu', 'numax', 'eps', 'd02', 'alpha', 'env_height',
                   'env_width', 'mode_width', 'teff', 'bp_rp']

        key: [log, log, lin, log, log, log, log, log, log, lin]
        '''
        # d02/dnu < 0.2  (np.log10(0.2) ~ -0.7)
        if p[3] - p[0] > -0.7:
            return -np.inf
        # Constraint from prior
        lp = np.log(self.kde.pdf(p))
        return lp

    def likelihood(self, p):
        '''
        Calculates the likelihood of the observed properties given
        the proposed parameters p.
        '''
        # log_dnu, log_numax, eps, log_d02, log_alpha, log_env_height, \
        #     log_env_width, log_mode_width, log_teff, bp_rp = p

        # Constraint from input data
        ld = 0.0
        ld += self.normal(p[0], *self.log_obs['dnu'])
        ld += self.normal(p[1], *self.log_obs['numax'])
        ld += self.normal(p[8], *self.log_obs['teff'])
        ld += self.normal(p[9], *self.log_obs['bp_rp'])
        return ld

    def kde_sampler(self, nwalkers=50):
        ''' Samples from the posterior probability distribution

        p(theta | D) propto p(theta) p(D | theta)

        p(theta) is given by the KDE function of the prior data

        p(D | theta) is given by the observable constraints

        Samples are drawn using `emcee`.

        Returns
        -------

        chains: flatchain
            The emcee chain of samples from the posterior

        Note:['dnu', 'numax', 'eps',
                     'd02', 'alpha', 'env_height',
                     'env_width', 'mode_width', 'teff',
                     'bp_rp']
        key: [log, log, lin, log, log, log, log, log, log, lin]

        '''
        if self.verbose:
            print('Running KDE sampler')

        x0 = [self.log_obs['dnu'][0],  # log10 dnu
              self.log_obs['numax'][0],  # log10 numax
              1.0,  # eps
              np.log10(0.1 * self.obs['dnu'][0]),  # log10 d02
              -2.0,  # log10 alpha
              1.0,  # log10 env height
              1.0,  # log10 env width,
              -1.0,  # log10 mode width
              self.log_obs['teff'][0],
              self.log_obs['bp_rp'][0]]

        self.fit = mcmc(x0, self.likelihood, self.prior, nwalkers=nwalkers)

        return self.fit()

    def to_log10(self, x, xerr):
        '''
        Transforms observables into log10 space.
        '''
        if xerr > 0:
            return [np.log10(x), xerr/x/np.log(10.0)]
        return [x, xerr]

    def obs_to_log(self, obs):
        '''
        Converted the observed properties into log10 space.
        '''
        self.log_obs = {'dnu': self.to_log10(*obs['dnu']),
                        'numax': self.to_log10(*obs['numax']),
                        'teff': self.to_log10(*obs['teff']),
                        'bp_rp': obs['bp_rp']}

    def kde_predict(self, n):
        '''
        Predict the frequencies from the kde method samples.

        The sampler must be run before calling the predict method.

        Parameters
        ----------

        n: numpy_array
            The radial order

        Returns
        -------
        frequencies: numpy-array
            A numpy array of length len(n) containting the frequency estimates.

        frequencies_unc: numpy-array
            A numpy array of length len(n) containting the frequency estimates
            uncertainties.
        '''
        if self.samples == []:
            print('Need to run the sampler first')
            return -1, -1
        dnu = 10**self.samples[:, 0]
        eps = self.samples[:, 2]
        nmax = 10**self.samples[:, 1] / dnu - eps
        alpha = 10**self.samples[:, 4]
        freq = np.array([(nn + eps + alpha/2.0 * (nn - nmax)**2) * dnu for nn in n])
        return freq.mean(axis=1), freq.std(axis=1)

    def __call__(self, dnu=[1, -1], numax=[1, -1], teff=[1, -1],
                 bp_rp=[1, -1]):
        ''' Calls the relevant defined method and returns an estimate of
        epsilon.

        Inputs
        ------
        dnu : [real, real]
            Large frequency spacing and uncertainty
        numax : [real, real]
            Frequency of maximum power and uncertainty
        teff : [real, real]
            Stellar effective temperature and uncertainty
        bp_rp : [real, real]
            The Gaia Gbp - Grp color value and uncertainty
            (probably ~< 0.01 dex)
        Returns
        -------
        result : array-like
            [estimate of epsilon, unceritainty on estimate]
        '''
        self.obs = {'dnu': dnu, 'numax': numax, 'teff': teff, 'bp_rp': bp_rp}
        
        self.obs_to_log(self.obs)

        self.make_kde()

        self.samples = self.kde_sampler()

        self.result = [self.samples[:, 2].mean(), self.samples[:, 2].std()]



import os
from .asy_peakbag import asymptotic_fit
from .priors import kde
from .peakbag import peakbag
from .jar import get_priorpath
from .plotting import plotting
import pandas as pd

class star(plotting):
    """ Class for each star to be peakbagged

    Additional attributes are added for each step of the peakbagging process

    Parameters
    ----------
    ID : string, int
        Target identifier. If custom timeseries/periodogram is provided, it
        must be resolvable by LightKurve (KIC, TIC, EPIC, HD, etc.).

    periodogram : lightkurve.periodogram.Periodogram object
        A lightkurve periodogram object containing frequencies in units of
        microhertz and power (in arbitrary units).

    numax : list
        List of the form [numax, numax_error]. For multiple targets, use a list
        of lists.

    dnu : list
        List of the form [dnu, dnu_error]. For multiple targets, use a list
        of lists.

    teff : list
        List of the form [teff, teff_error]. For multiple targets, use a list
        of lists.

    bp_rp : list
        List of the form [bp_rp, bp_rp_error]. For multiple targets, use a list
        of lists.

    store_chains : bool, optional
        Flag for storing all the full set of samples from the MCMC run.
        Warning, if running multiple targets, make sure you have enough memory.

    nthreads : int, optional
        Number of multiprocessing threads to use to perform the fit. For long
        cadence data 1 is best, more will just add parallelization overhead.
        Untested on short cadence.

    make_plots : bool, optional
        If True, will save figures when calling methods in `star`.

    path : str, optional
        The path at which to store output. If no path is set but make_plots is
        True, output will be saved in the current working directory.

    verbose : bool, optional
        If True, will show error messages on the users terminal if they occur.
        
    Attributes
    ----------
    f : array
        Array of power spectrum frequencies
    s : array
        power spectrum
    data_file : str
        Path to the csv file containing the prior data
    """

    def __init__(self, ID, pg, numax, dnu, teff, bp_rp, path=None, 
                 prior_file = None):
          
        self.ID = ID
        self.pg = pg
        self.f = pg.frequency.value
        self.s = pg.power.value

        self.numax = numax
        self.dnu = dnu
        self.teff = teff
        self.bp_rp = bp_rp

        self.make_output_dir(path)

        if not prior_file:
            self.prior_file = get_priorpath() 
        else:
            self.prior_file = prior_file
    
    def make_output_dir(self, path):
        
        # Check if self has path attribute
        if not hasattr(self, 'path'):
            self.path = None
            
        if isinstance(self.path, str):  # If yes, do nothing
            pass
        else:  # If not, assume path should be cwd+ID
            self.path = os.path.join(*[os.getcwd(), f'{self.ID}'])
        
        # If path is str, presume user wants to override self.path
        if isinstance(path, str):
            self.path =os.path.join(*[path, f'{self.ID}'])

        # Check if self.path exists, if not try to create it
        if not os.path.isdir(self.path):
            try:
                os.makedirs(self.path)
            except Exception as ex:
                message = "Star {0} produced an exception of type {1} occurred. Arguments:\n{2!r}".format(self.ID, type(ex).__name__, ex.args)
                print(message)
          
    def run_kde(self, bw_fac=1.0, make_plots=False):
        """
        Runs the kde code and makes plots if self.make_plots is set.
        """
        print('Starting KDE estimation')
        # Init
        kde(self, bw_fac=bw_fac)
        
        # Call
        self.kde(dnu=self.dnu, numax=self.numax, teff=self.teff, 
                 bp_rp=self.bp_rp)
        
        # Store
        if make_plots:
            self.kde.plot_corner(path=self.path, ID=self.ID, 
                                 savefig=make_plots)
            self.kde.plot_spectrum(pg=self.pg, path=self.path, ID=self.ID, 
                                       savefig=make_plots)
             
            
    def run_asy_peakbag(self, norders=None, make_plots=False, 
                        store_chains=False, nthreads=1):
        """
        Runs the asy_peakbag code.
        """
        print('Starting Asy_peakbag')
        # Init
        asymptotic_fit(self, self.kde, norders=norders, 
                       store_chains=store_chains, nthreads=nthreads)
        
        # Call
        self.asy_fit(dnu=self.dnu, numax=self.numax, teff=self.teff, 
                     bp_rp=self.bp_rp)
        
        # Store
        outpath = lambda x: os.path.join(*[self.path, x])
        self.asy_fit.summary.to_csv(outpath(f'asy_fit_summary_{self.ID}.csv'),
                                    index=True)
        self.asy_fit.modeID.to_csv(outpath(f'asy_fit_modeID_{self.ID}.csv'),
                                   index=False)
        
        if make_plots:
            self.asy_fit.plot_spectrum(path=self.path, ID=self.ID, 
                                       savefig=make_plots)
            self.asy_fit.plot_corner(path=self.path, ID=self.ID, 
                                       savefig=make_plots)
        
        if store_chains:
            pd.DataFrame(self.asy_fit.samples, columns=self.asy_fit.par_names).to_csv(outpath(f'asy_peakbag_chains_{self.ID}.csv'), index=False) 
        
            

    def run_peakbag(self, model_type='simple', tune=1500, nthreads=1, make_plots=False, store_chains=False):
        """
        Runs peakbag on the given star.
        """
        
        print('Starting peakbagging run')
        # Init
        self.peakbag = peakbag(self, self.asy_fit)
        
        # Call
        self.peakbag(model_type=model_type, tune=tune, nthreads=nthreads)
        
        # Store
        outpath = lambda x: os.path.join(*[self.path, x])
        self.peakbag.summary.to_csv(outpath(f'peakbag_summary_{self.ID}.csv'))
        
        if store_chains:
            pass # TODO need to pickle the samples if requested.
        if make_plots:
            self.peakbag.plot_spectrum(path=self.path, ID=self.ID, 
                                       savefig=make_plots)


    def __call__(self, bw_fac=1.0, norders=8, model_type='simple', tune=1500, 
                 verbose=False, make_plots=True, store_chains=True, nthreads=1):
        """ Instead of a _call_ we should just make this a function maybe? Whats wrong with __call__?"""
        
        self.run_kde(bw_fac=bw_fac, make_plots=make_plots)          
   
        self.run_asy_peakbag(norders=norders, make_plots=make_plots, 
                             store_chains=store_chains, nthreads=nthreads)
        
        self.run_peakbag(model_type=model_type, tune=tune, nthreads=nthreads, 
                         make_plots=make_plots, store_chains=store_chains)



import matplotlib.pyplot as plt
import astropy.convolution as conv
import pbjam, os, corner
import numpy as np
from pymc3.gp.util import plot_gp_dist
import astropy.units as u
import pymc3 as pm

class plotting():
    
    def __init__(self):
        pass
    
    def plot_echelle(self, pg=None):
        '''
        Plots an echelle diagram with mode frequencies over plotted.
    
        Parameters
        ----------
        pg : periodogram
            A lightkurve periodogram (we use the plot echelle method
            of the periodogram (well actually seismology) class)
    
        Returns
        -------
        fig : figure
            The figure containing the plot.
        '''
                
        freqs = {'l'+str(i): {} for i in range(4)}
        
        if type(self) == pbjam.star:
            dnu = self.dnu[0]
            numax = self.numax[0]
            
        elif type(self) == pbjam.priors.kde:  
            dnu = 10**np.median(self.samples[:,0])
            numax = 10**np.median(self.samples[:,1])
            
        elif type(self) == pbjam.asy_peakbag.asymptotic_fit:
            dnu = 10**self.summary.loc['dnu', '50th']
            numax = 10**self.summary.loc['numax', '50th']
            
            for l in np.arange(4):
                idx = self.modeID.ell == l
                freqs['l'+str(l)] = {'nu': self.modeID.loc[idx, 'nu_med'],
                                     'err': self.modeID.loc[idx, 'nu_mad']}
                
        elif type(self) == pbjam.peakbag:  
            numax = 10**self.asy_result.summary.loc['numax', '50th']
            for l in np.arange(4):
                ell = 'l'+str(l)
                freqs[ell] = {'nu': self.summary.filter(like=ell, axis=0).loc[:, 'mean'],
                              'err': self.summary.filter(like=ell, axis=0).loc[:, 'sd']}
            dnu = np.median(np.diff(freqs['l0']['nu']))
            
        else:
            raise ValueError('Unrecognized class type')
            
        # make dnu an intger multiple of bw
        dnu -= dnu % (self.f[1] - self.f[0])
        nmin = np.floor(self.f.min() / dnu) + 1
        
        if pg:
            peri = pg
        elif hasattr(self, 'pg'):
            peri = self.pg
        else:
            raise ValueError('Need spectrum to plot echelle diagram')
            
        seismology = peri.flatten().to_seismology()
        
        ax = seismology.plot_echelle(deltanu=dnu * u.uHz,
                                     numax=numax * u.uHz,
                                     minimum_frequency=dnu*nmin)

        cols = ['C0', 'C1', 'C2', 'C3']
        print(freqs)
        for l in np.arange(4):
            ell = 'l'+str(l)
            if len(freqs[ell].nu) > 0:
                nu = freqs[ell]['nu']
                err = freqs[ell]['err']
                ax.errorbar(nu%dnu, (nu//dnu) * dnu, xerr=err, fmt='o', color = cols[l], label = r'$\ell=$%i' % (l))
        ax.legend(fontsize = 'x-small')
        
    def plot_corner(self, path=None, ID=None, savefig=False):
       
        '''
        Makes a nice corner plot using corner.corner
        '''
        
        fig = corner.corner(self.samples, labels=self.par_names,
                            show_titles=True, quantiles=[0.16, 0.5, 0.84],
                            title_kwargs={"fontsize": 12})
        
        # TODO there should be a check if path is full filepath or just dir
        if path and ID:
            outpath = os.path.join(*[path,  type(self).__name__+ '_corner_' + str(ID) + '.png'])
            if savefig:
                fig.savefig(outpath)
                
        return fig


    def plot_spectrum(self, pg=None, path=None, ID=None, savefig=False):
    
        if not pg and hasattr(self, 'pg'):
            pg = self.pg
        
        if pg:
            f = pg.frequency.value
            s = pg.power.value
            
        elif hasattr(self, 'f') & hasattr(self, 's'):
            f = self.f
            s = self.s            
        else:
            raise ValueError('Unable to plot spectrum.')
            
        # The raw and smoothed spectrum will always be plotted
        fig, ax = plt.subplots(figsize=[16,9])
        ax.plot(f, s, 'k-', label='Data', alpha=0.2)
        fac = 0.1 / (f[1] - f[0]) #0.005 * self.dnu[0]  / (f[1] - f[0])
        kernel = conv.Gaussian1DKernel(stddev=fac)
        smoo = conv.convolve(s, kernel)
        ax.plot(f, smoo, 'k-', label='Smoothed', lw=3, alpha=0.6)
    
        # Overplot kde diagnostic  
        if type(self) == pbjam.star:
             xlim = [self.numax[0]-5*self.dnu[0], self.numax[0]+5*self.dnu[0]]
         
        elif type(self) == pbjam.priors.kde:
            h = max(smoo)
            dnu = 10**(np.median(self.samples[:, 0]))
            nmin = np.floor(min(f) / dnu)
            nmax = np.floor(max(f) / dnu)
            enns = np.arange(nmin-1, nmax+1, 1)
            freq, freq_sigma = self.kde_predict(enns)
            y = np.zeros(len(f))
            for i in range(len(enns)):
                y += 0.8 * h * np.exp(-0.5 * (freq[i] - f)**2 / freq_sigma[i]**2)
            ax.fill_between(f, y, alpha=0.3, facecolor='navy', edgecolor='none',
                            label=r'$\propto P(\nu_{\ell=0})$')

            xlim = [min(freq)-dnu, max(freq)+dnu]
            
        # Overplot asy_peakbag diagnostic        
        elif type(self) == pbjam.asy_peakbag.asymptotic_fit:
            for j in np.arange(-50,0):
                if j==-1:
                    label='Model'
                else:
                    label=None
                ax.plot(f[self.sel], self.model(self.samples[j, :]), 'r-', 
                        alpha=0.1)
            for nu in self.modeID['nu_med']:
                ax.axvline(nu, c='k', linestyle='--')
            dnu = 10**self.summary.loc['dnu', '50th']
            xlim = [min(f[self.sel])-dnu, 
                    max(f[self.sel])+dnu]
            
        # Overplot peakbag diagnostic    
        elif type(self) == pbjam.peakbag:  
            n = self.ladder_s.shape[0]                
            par_names = ['l0', 'l2', 'width0', 'width2', 'height0', 'height2', 
                         'back']
            for i in range(n):
                for j in range(-50, 0):
                    if (i == 0) and (j==-1):
                        label='Model'
                    else:
                        label=None
                    mod = self.model(*[self.samples[x][j] for x in par_names])
                    ax.plot(self.ladder_f[i, :], mod[i, :], c='r', alpha=0.1, 
                            label=label)
                    
            dnu = 10**self.asy_result.summary.loc['dnu', '50th']        
            xlim = [min(f[self.asy_result.sel])-dnu, 
                    max(f[self.asy_result.sel])+dnu]
        else:
            raise ValueError('Unrecognized class type')

        ax.set_ylim([0, smoo.max()*1.5])
        ax.set_xlim([max([min(f), xlim[0]]), min([max(f), xlim[1]])])
        ax.set_xlabel(r'Frequency ($\mu \rm Hz$)')
        ax.set_ylabel(r'SNR')
        ax.legend(loc=1)
    
        if savefig:
            outpath = os.path.join(*[path, f'{type(self).__name__}_{str(ID)}.png'])
            fig.savefig(outpath)


    def plot_trace(stage):
        '''
        Will make a pymc3 traceplot.
        '''
        import pymc3 as pm
        
        if type(stage) == pbjam.priors.kde:
            # TODO - make this work for kde
            print('Traceplot for kde not yet implimented')
        
        if type(stage) == pbjam.asy_peakbag.asymptotic_fit:
            # TODO - make this work for asy_peakbag
            print('Traceplot for asy_peakbag not yet implimented')
        
        if type(stage) == pbjam.peakbag:
            pm.traceplot(stage.samples)


# Asy_peakbag  
    def plot_start(self):
        '''
        Plots the starting model as a diagnotstic.
        '''
             
        dnu = 10**np.median(self.start_samples, axis=0)[0]
        xlim = [min(self.f[self.sel])-dnu, max(self.f[self.sel])+dnu]
        fig, ax = plt.subplots(figsize=[16,9])
        ax.plot(self.f, self.s, 'k-', label='Data', alpha=0.2)
        smoo = dnu * 0.005 / (self.f[1] - self.f[0])
        kernel = conv.Gaussian1DKernel(stddev=smoo)
        smoothed = conv.convolve(self.s, kernel)
        ax.plot(self.f, smoothed, 'k-', label='Smoothed', lw=3, alpha=0.6)
        ax.plot(self.f[self.sel], self.model(self.start_samples.mean(axis=0)), 
                'r-', label='Start model', alpha=0.7)
        ax.set_ylim([0, smoothed.max()*1.5])
        ax.set_xlim(xlim)
        ax.set_xlabel(r'Frequency ($\mu \rm Hz$)')
        ax.set_ylabel(r'SNR')
        ax.legend()
        return fig


 
      
# Peakbag
def plot_linewidth(self, thin=10):
    """
    Plots the estimated line width as a function of scaled n.
    """
    fig, ax = plt.subplots(1, 2, figsize=[16,9])

    if self.gp0 != []:

        n_new = np.linspace(-0.2, 1.2, 100)[:,None]
        with self.pm_model:
            f_pred0 = self.gp0.conditional("f_pred0", n_new)
            f_pred2 = self.gp2.conditional("f_pred2", n_new)
            self.pred_samples = pm.sample_posterior_predictive(self.samples,
                           vars=[f_pred0, f_pred2], samples=1000)
        plot_gp_dist(ax[0], self.pred_samples["f_pred0"], n_new)
        plot_gp_dist(ax[1], self.pred_samples["f_pred2"], n_new)

        for i in range(0, len(self.samples), thin):
            ax[0].scatter(self.n,
                          self.samples['ln_width0'][i, :], c='k', alpha=0.3)
            ax[1].scatter(self.n,
                          self.samples['ln_width2'][i, :], c='k', alpha=0.3)


    else:
        for i in range(0, len(self.samples), thin):
            ax[0].scatter(self.n,
                          np.log(self.samples['width0'][i, :]), c='k', alpha=0.3)
            ax[1].scatter(self.n,
                          np.log(self.samples['width2'][i, :]), c='k', alpha=0.3)

    ax[0].set_xlabel('normalised order')
    ax[1].set_xlabel('normalised order')
    ax[0].set_ylabel('ln line width')
    ax[1].set_ylabel('ln line width')
    ax[0].set_title('Radial modes')
    ax[1].set_title('Quadrupole modes')
    return fig

def plot_height(self, thin=10):
    """
    Plots the estimated mode height.
    """
    fig, ax = plt.subplots(figsize=[16,9])
    for i in range(0, len(self.samples), thin):
        ax.scatter(self.samples['l0'][i, :], self.samples['height0'][i, :])
        ax.scatter(self.samples['l2'][i, :], self.samples['height2'][i, :])
    return fig

def plot_ladder(self, thin=10, alpha=0.2):
    """
    Plots the ladder data and models from the samples

    Parameters
    ----------
    thin: int
        Uses every other thin'th value from the samkles, i.e. [::thin].
    alpha: float64
        The alpha to use for plotting the models from samples.
    """
    n = self.ladder_s.shape[0]
    fig, ax = plt.subplots(n, figsize=[16,9])
    for i in range(n):
        for j in range(0, len(self.samples), thin):
            mod = self.model(self.samples['l0'][j],
                             self.samples['l2'][j],
                             self.samples['width0'][j],
                             self.samples['width2'][j],
                             self.samples['height0'][j],
                             self.samples['height2'][j],
                             self.samples['back'][j])
            ax[i].plot(self.ladder_f[i, :], mod[i, :], c='r', alpha=alpha)
        ax[i].plot(self.ladder_f[i, :], self.ladder_s[i, :], c='k')
        ax[i].set_xlim([self.ladder_f[i, 0], self.ladder_f[i, -1]])
    ax[n-1].set_xlabel(r'Frequency ($\mu \rm Hz$)')
    fig.tight_layout()
    return fig



